import os
import numpy as np
import math
import six
import iris
import iris.cube
from iris.cube import CubeList
import iris.analysis
import iris.util
import iris.coords
import warnings
import numba
import datetime
from numba import vectorize, guvectorize
from numba import cuda
from numba import int32, int64, float32, float64
from numba.cuda.cudadrv import driver

MESH_FILE = "/esarchive/autosubmit/con_files/mesh_mask_nemo.Ec3.2_O25L75.nc"
REGIONS_FILE = "/esarchive/autosubmit/con_files/mask.regions.Ec3.2_O25L75.nc"
VO_FILE = "/esarchive/exp/ecearth/a0qp/cmorfiles/CMIP/EC-Earth-Consortium/" \
          "EC-Earth3-HR/historical/r1i1p1f1/Omon/vo/gn/v20190411/" \
          "vo_Omon_EC-Earth3-HR_historical_r1i1p1f1_gn_201301-201312.nc"
RES = '/esarchive/exp/ecearth/a16l/cmorfiles/CMIP/EC-Earth-Consortium' \
      '/EC-Earth3-LR/historical/r1i1p1f1/Omon/vsftmyz/gn/v20180711' \
      '/vsftmyz_Omon_EC-Earth3-LR_historical_r1i1p1f1_gn_196101-196112.nc'


def main():
    e3v, e1v = load_mesh()
    basin_name, basins = load_masks()
    area = {}
    for basin in basins:
        area[basin] = compute_area_basin(e1v, e3v, basins[basin])
    del e3v, e1v
    vo = load_vo()
    for i in range(1):
        start = datetime.datetime.now()
        vsftmyz = compute_vsftmyz_cpu(vo, area)
        print('np {0}'.format(datetime.datetime.now()-start))
        start = datetime.datetime.now()
        moc = compute_vsftmyz_gpu(vo, area)
        print('gp {0}'.format(datetime.datetime.now()-start))
    save_data(moc)
    save_data_cpu(vsftmyz)


def compute_vsftmyz_cpu(vo, area):
    vsftmyz = []
    for basin in area:
        vsftmyz_basin = np.sum(multiply_vo_basin(vo, area[basin]), axis=3)
        compute_moc(vsftmyz_basin, vsftmyz_basin)
        vsftmyz.append(vsftmyz_basin)
    return vsftmyz


def compute_vsftmyz_gpu(vo, area):
    times = vo.shape[0]
    levels = vo.shape[1]
    lats = vo.shape[2]
    lons = vo.shape[3]

    block = (128, 1, 1)
    grid_size = (lats // block[0]) + 1
    grid3D = (grid_size, levels, times)
    gpu_vo = cuda.to_device(vo.astype(np.float32))
    gpu_moc = cuda.device_array((times, levels, lats), dtype=np.float32)
    grid2D = (grid_size, times)

    vsftmyz = []
    for basin in area:
        gpu_area = cuda.to_device(area[basin].astype(np.float32))
        horizontal_integral[grid3D, block](gpu_vo, gpu_area, gpu_moc)
        compute_moc_gpu[grid2D, block](gpu_moc)
        vsftmyz.append(gpu_moc.copy_to_host())

    return vsftmyz


def load_vo():
    vo = iris.load_cube(VO_FILE, 'sea_water_y_velocity')
    vo_data = np.ma.filled(vo.data, 0.0).astype(np.float32)
    return vo_data


def load_mesh():
    e3v = iris.load_cube(MESH_FILE, 'e3v_0').data.astype(np.float32)
    e1v = iris.load_cube(MESH_FILE, 'e1v').data.astype(np.float32)
    return e3v, e1v


def load_masks():
    basins = {}
    global_mask = iris.load_cube(MESH_FILE, 'vmask')
    basin_name = []
    basin_name.append('Global_Ocean')
    global_mask.data[..., 0] = 0.0
    global_mask.data[..., -1] = 0.0
    basins['Global_Ocean'] = global_mask.data.astype(np.float32)
    cubes = iris.load(REGIONS_FILE)
    for cube in cubes:
        name = cube.name()
        if name in ('nav_lat', 'nav_lon', 'Caspian_Sea', 'Global_Ocean'):
            continue
        cube = cube.extract(iris.Constraint(z=1))
        basins[name] = cube.data.astype(np.float32)
    return basin_name, basins


def save_data(vsftmyz):
    moc_cube = []
    for basin in range(len(vsftmyz)):
        moc_cube.append(iris.cube.Cube(vsftmyz[basin]))
    iris.save(moc_cube, '/esarchive/scratch/sloosvel/numba_outputs/moc.nc',
              zlib=True)


def save_data_cpu(vsftmyz):
    moc_cube = []
    for basin in range(len(vsftmyz)):
        moc_cube.append(iris.cube.Cube(vsftmyz[basin]))
    iris.save(moc_cube, '/esarchive/scratch/sloosvel/numba_outputs/vsftmyz.nc',
              zlib=True)


@vectorize(['float32(float32, float32, float32)'], target='cpu')
def compute_area_basin(e1v, e3v, basin):
    return - e1v * e3v * basin / 1e6


@vectorize(['float32(float32, float32)'], target='cpu')
def multiply_vo_basin(vo, basin):
    return vo*basin


@guvectorize([(float32[:, :, :], float32[:, :, :])], '(t, l, j)->(t, l, j)',
             target='cpu')
def compute_moc(moc, out):
    for lev in reversed(range(moc.shape[1]-1)):
        moc[:, lev, :] = moc[:, lev, :] + moc[:, lev+1, :]


@cuda.jit()
def compute_moc_gpu(moc):
    j, t = cuda.grid(2)
    if(j >= moc.shape[2]):
        return
    for lev in range(moc.shape[1]-2, -1, -1):
        moc[t, lev, j] = moc[t, lev, j] + moc[t, lev+1, j]


@cuda.jit()
def horizontal_integral(vo, area, moc):
    j, lev, t = cuda.grid(3)
    moc[t, lev, j] = 0.0
    temp = 0.0
    if(j >= vo.shape[2]):
        return
    for i in range(vo.shape[3]):
        temp += vo[t, lev, j, i] * area[0, lev, j, i]
    moc[t, lev, j] = temp


if __name__ == '__main__':
    main()
